import os,getpass,time
from dpdispatcher.Batch import Batch
from dpdispatcher.JobStatus import JobStatus

def _default_item(resources, key, value) :
    if key not in resources :
        resources[key] = value

class Slurm(Batch) :

    def check_status(self) :
        """
        check the status of a job
        """
        job_id = self._get_job_id()
        if job_id == '' :
            return JobStatus.unsubmitted
        while True:
            stat = self._check_status_inner(job_id)
            if stat != JobStatus.completing:
                return stat
            else:
                time.sleep(5)

    def do_submit(self, 
                  job_dirs,
                  cmd,
                  args = None, 
                  res = None,
                  outlog = 'log',
                  errlog = 'err'):
        if res == None:
            res = self.default_resources(res)
        if 'task_max' in res and res['task_max'] > 0:
            while self._check_sub_limit(task_max=res['task_max']):
                time.sleep(60)
        script_str = self.sub_script(job_dirs, cmd, args=args, res=res, outlog=outlog, errlog=errlog)
        self.context.write_file(self.sub_script_name, script_str)
        stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'sbatch', self.sub_script_name))
        subret = (stdout.readlines())
        job_id = subret[0].split()[-1]
        self.context.write_file(self.job_id_name, job_id)        
                
    def default_resources(self, res_) :
        """
        set default value if a key in res_ is not fhound
        """
        if res_ == None :
            res = {}
        else:
            res = res_
        _default_item(res, 'numb_node', 1)
        _default_item(res, 'task_per_node', 1)
        _default_item(res, 'cpus_per_task', -1)
        _default_item(res, 'numb_gpu', 0)
        _default_item(res, 'time_limit', '1:0:0')
        _default_item(res, 'mem_limit', -1)
        _default_item(res, 'partition', '')
        _default_item(res, 'account', '')
        _default_item(res, 'qos', '')
        _default_item(res, 'constraint_list', [])
        _default_item(res, 'license_list', [])
        _default_item(res, 'exclude_list', [])
        _default_item(res, 'module_unload_list', [])
        _default_item(res, 'module_list', [])
        _default_item(res, 'source_list', [])
        _default_item(res, 'envs', None)
        _default_item(res, 'with_mpi', False)
        _default_item(res, 'cuda_multi_tasks', False)
        _default_item(res, 'allow_failure', False)
        _default_item(res, 'cvasp', False)
        return res

    def sub_script_head(self, res):
        ret = ''
        ret += "#!/bin/bash -l\n"
        ret += "#SBATCH -N %d\n" % res['numb_node']
        ret += "#SBATCH --ntasks-per-node=%d\n" % res['task_per_node']
        if res['cpus_per_task'] > 0 :            
            ret += "#SBATCH --cpus-per-task=%d\n" % res['cpus_per_task']
        ret += "#SBATCH -t %s\n" % res['time_limit']
        if res['mem_limit'] > 0 :
            ret += "#SBATCH --mem=%dG \n" % res['mem_limit']
        if len(res['account']) > 0 :
            ret += "#SBATCH --account=%s \n" % res['account']
        if len(res['partition']) > 0 :
            ret += "#SBATCH --partition=%s \n" % res['partition']
        if len(res['qos']) > 0 :
            ret += "#SBATCH --qos=%s \n" % res['qos']
        if res['numb_gpu'] > 0 :
            ret += "#SBATCH --gres=gpu:%d\n" % res['numb_gpu']
        for ii in res['constraint_list'] :
            ret += '#SBATCH -C %s \n' % ii
        for ii in res['license_list'] :
            ret += '#SBATCH -L %s \n' % ii
        if len(res['exclude_list']) >0:
            temp_exclude = ""
            for ii in res['exclude_list'] :
                temp_exclude += ii
                temp_exclude += ","
            temp_exclude = temp_exclude[:-1]
            ret += '#SBATCH --exclude=%s \n' % temp_exclude
        ret += "\n"
        for ii in res['module_unload_list'] :
            ret += "module unload %s\n" % ii
        for ii in res['module_list'] :
            ret += "module load %s\n" % ii
        ret += "\n"
        for ii in res['source_list'] :
            ret += "source %s\n" %ii
        ret += "\n"
        envs = res['envs']
        if envs != None :
            for key in envs.keys() :
                ret += 'export %s=%s\n' % (key, envs[key])
            ret += '\n'        
        return ret

    def sub_script_cmd(self,
                       cmd,
                       arg,
                       res) :
        try:
            cvasp=res['cvasp']
            fp_max_errors = 3
            try:
                fp_max_errors = res['fp_max_errors']
            except:
                pass
        except:
            cvasp=False

        _cmd = cmd.split('1>')[0].strip()
        if cvasp :
            if res['with_mpi']:
                _cmd = 'python cvasp.py "srun %s %s" %s' % (_cmd, arg, fp_max_errors)
            else :
                _cmd = 'python cvasp.py "%s %s" %s' % (_cmd, arg, fp_max_errors)
        else :
            if res['with_mpi']:
                _cmd = 'srun %s %s' % (_cmd, arg)
            else :
                _cmd = '%s %s' % (_cmd, arg)        
        return _cmd

    def _get_job_id(self) :
        if self.context.check_file_exists(self.job_id_name) :
            return self.context.read_file(self.job_id_name)
        else:
            return ""

    def _check_status_inner(self, job_id):
        ret, stdin, stdout, stderr\
            = self.context.block_call ('squeue -o "%.18i %.2t" -j ' + job_id)
        if (ret != 0) :
            err_str = stderr.read().decode('utf-8')
            if str("Invalid job id specified") in err_str :
                if self.check_finish_tag() :
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            else :
                raise RuntimeError\
                    ("status command squeue fails to execute\nerror message:%s\nreturn code %d\n" % (err_str, ret))
        status_line = stdout.read().decode('utf-8').split ('\n')[-2]
        status_word = status_line.split ()[-1]
        if not (len(status_line.split()) == 2 and status_word.isupper()): 
            raise RuntimeError("Error in getting job status, " +
                              f"status_line = {status_line}, " + 
                              f"parsed status_word = {status_word}")
        if status_word in ["PD","CF","S"] :
            return JobStatus.waiting
        elif status_word in ["R"] :
            return JobStatus.running
        elif status_word in ["CG"] :
            return JobStatus.completing
        elif status_word in ["C","E","K","BF","CA","CD","F","NF","PR","SE","ST","TO"] :
            if self.check_finish_tag() :
                return JobStatus.finished
            else :
                return JobStatus.terminated
        else :
            return JobStatus.unknown                    


    def _check_sub_limit(self, task_max, **kwarg) :
        if task_max <= 0:
            return True
        username = getpass.getuser()
        stdin, stdout, stderr = self.context.block_checkcall('squeue -u %s -h' % username)
        nj = len(stdout.readlines())
        return nj >= task_max

    def _make_squeue(self,mdata1, res):
        ret = ''
        ret += 'squeue -u %s ' % mdata1['username']
        ret += '-p %s ' % res['partition']
        ret += '| grep PD'
        return ret



#!/usr/bin/env python
# coding: utf-8

from dpdispatcher.base_context import BaseContext
import os, sys, paramiko, json, uuid, tarfile, time, stat, shutil
from glob import glob
from dpdispatcher import dlog
# from dpdispatcher.submission import Machine

class SSHSession (object):
    def __init__(self,
                hostname,
                username,
                password=None,
                port=22,
                key_filename=None,
                passphrase=None,
                timeout=10):

        self.hostname = hostname
        self.username = username
        self.password = password
        self.port = port
        self.key_filename = key_filename
        self.passphrase = passphrase
        self.timeout = timeout
        self.ssh = None
        self._setup_ssh()

    # @classmethod
    # def deserialize(cls, jdata):
    #     instance = cls(**jdata)
    #     return instance
        
    # def bk_ensure_alive(self,
    #                  max_check = 10,
    #                  sleep_time = 10):
    #     count = 1
    #     while not self._check_alive():
    #         if count == max_check:
    #             raise RuntimeError('cannot connect ssh after %d failures at interval %d s' %
    #                                (max_check, sleep_time))
    #         dlog.info('connection check failed, try to reconnect to ' + self.remote_host)
    #         self._setup_ssh(hostname=self.remote_host,
    #                         port=self.remote_port,
    #                         username=self.remote_uname,
    #                         password=self.remote_password,
    #                         key_filename=self.local_key_filename,
    #                         timeout=self.remote_timeout,
    #                         passphrase=self.local_key_passphrase)
    #         count += 1
    #         time.sleep(sleep_time)

    def ensure_alive(self,
                    max_check = 10,
                    sleep_time = 10):
        count = 1
        while not self._check_alive():
            if count == max_check:
                raise RuntimeError('cannot connect ssh after %d failures at interval %d s' %
                                    (max_check, sleep_time))
            dlog.info('connection check failed, try to reconnect to ' + self.remote_root)
            self._setup_ssh()
            count += 1
            time.sleep(sleep_time)

    def _check_alive(self):
        if self.ssh == None:
            return False
        try :
            transport = self.ssh.get_transport()
            transport.send_ignore()
            return True
        except EOFError:
            return False        

    # def bk_setup_ssh(self,
    #                hostname,
    #                port=22,
    #                username=None,
    #                password=None,
    #                key_filename=None,
    #                timeout=None,
    #                passphrase=None):
    #     self.ssh = paramiko.SSHClient()
    #     # ssh_client.load_system_host_keys()
    #     self.ssh.set_missing_host_key_policy(paramiko.WarningPolicy)
    #     self.ssh.connect(hostname=hostname, port=port,
    #                      username=username, password=password,
    #                      key_filename=key_filename, timeout=timeout, passphrase=passphrase)
    #     assert(self.ssh.get_transport().is_active())
    #     transport = self.ssh.get_transport()
    #     transport.set_keepalive(60)
    
    def _setup_ssh(self):
        # machine = self.machine
        self.ssh = paramiko.SSHClient()
        self.ssh.set_missing_host_key_policy(paramiko.WarningPolicy)
        self.ssh.connect(hostname=self.hostname, port=self.port,
                        username=self.username, password=self.password,
                        key_filename=self.key_filename, timeout=self.timeout,passphrase=self.passphrase)
        assert(self.ssh.get_transport().is_active())
        transport = self.ssh.get_transport()
        transport.set_keepalive(60)
        # reset sftp
        self._sftp = None
                        

    def get_ssh_client(self) :
        return self.ssh

    # def get_session_root(self) :
    #     return self.remote_root

    def close(self) :
        self.ssh.close()

    def exec_command(self, cmd, retry = 0):
        """Calling self.ssh.exec_command but has an exception check."""
        try:
            return self.ssh.exec_command(cmd)
        except paramiko.ssh_exception.SSHException:
            # SSH session not active
            # retry for up to 3 times
            if retry < 3:
                dlog.warning("SSH session not active in calling %s, retry the command..." % cmd)
                # ensure alive
                self.ensure_alive()
                return self.exec_command(cmd, retry = retry+1)
            raise RuntimeError("SSH session not active")

    @property
    def sftp(self):
        """Returns sftp. Open a new one if not existing."""
        if self._sftp is None:
            self.ensure_alive()
            self._sftp = self.ssh.open_sftp()
        return self._sftp


class SSHContext(BaseContext):
    def __init__ (self,
                local_root,
                remote_root,
                remote_profile,
                clean_asynchronously=False,
                ):
        assert(type(local_root) == str)
        self.temp_local_root = os.path.abspath(local_root)
        # self.job_uuid = None
        self.clean_asynchronously = clean_asynchronously
        # self.job_uuid = job_uuid
        # if job_uuid:
        #    self.job_uuid=job_uuid
        # else:
        #    self.job_uuid = str(uuid.uuid4())
        self.ssh_session = SSHSession(**remote_profile)
        # self.temp_remote_root = os.path.join(self.ssh_session.get_session_root())
        self.temp_remote_root = remote_root
        self.ssh_session.ensure_alive()
        try:
            self.sftp.mkdir(self.temp_remote_root)
        except OSError: 
            pass

    @classmethod
    def load_from_dict(cls, context_dict):
        # instance = cls()
        # input = dict(
        #     hostname = jdata['hostname'],
        #     remote_root = jdata['remote_root'],
        #     username = jdata['username'],
        #     password = jdata.get('password', None),
        #     port = jdata.get('port', 22),
        #     key_filename = jdata.get('key_filename', None),
        #     passphrase = jdata.get('passphrase', None),
        #     timeout = jdata.get('timeout', 10),
        # )
        local_root = context_dict['local_root']
        remote_root = context_dict['remote_root']
        remote_profile = context_dict['remote_profile']
        clean_asynchronously = context_dict.get('clean_asynchronously', False)

        ssh_context = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile,
            clean_asynchronously=clean_asynchronously
        )
        # local_root = jdata['local_root']
        # ssh_session = SSHSession(**input)
        # ssh_context = SSHContext(
        #     local_root=local_root,
        #     ssh_session=ssh_session,
        #     clean_asynchronously=jdata.get('clean_asynchronously', False),
        #     )
        return ssh_context
    
    @property
    def ssh(self):
        return self.ssh_session.get_ssh_client()  

    @property
    def sftp(self):
        return self.ssh_session.sftp

    def close(self):
        self.ssh_session.close()

    def get_job_root(self) :
        return self.remote_root

    def bind_submission(self, submission):
        self.submission = submission
        self.local_root = os.path.join(self.temp_local_root, submission.work_base)
        # self.remote_root = os.path.join(self.temp_remote_root, self.submission.submission_hash, self.submission.work_base )
        self.remote_root = os.path.join(self.temp_remote_root, self.submission.submission_hash)

        # self.job_uuid = submission.submission_hash
        print('debug:SSHContext.bind_submission', 
            submission.submission_hash,
            self.local_root, self.remote_root)

        # try:
        # print('self.remote_root', self.remote_root)
        # sftp = self.ssh_session.ssh.open_sftp() 
        # sftp.mkdir(self.remote_root)
        # sftp.close()
        # except:
        #     pass
        
    def upload(self,
               # job_dirs,
               submission,
               # local_up_files,
               dereference = True) :
        self.ssh_session.ensure_alive()
        cwd = os.getcwd()
        os.chdir(self.local_root) 
        file_list = []
        
      #   for ii in job_dirs :
        for task in submission.belonging_tasks :
            for jj in task.forward_files :
                # file_list.append(os.path.join(ii, jj))        
                file_list.append(os.path.join(task.task_work_path, jj))        
        # for ii in submission.forward_common_files:
        #     file_list.append(ii)
        file_list.extend(submission.forward_common_files)

        self._put_files(file_list, dereference = dereference)
        os.chdir(cwd)

    def download(self, 
                 submission,
                 # job_dirs,
                 # remote_down_files,
                 check_exists = False,
                 mark_failure = True,
                 back_error=False) :
        self.ssh_session.ensure_alive()
        cwd = os.getcwd()
        os.chdir(self.local_root) 
        file_list = []
        # for ii in job_dirs :
        for task in submission.belonging_tasks :
            for jj in task.backward_files:
                file_name = os.path.join(task.task_work_path, jj)                
                if check_exists:
                    if self.check_file_exists(file_name):
                        file_list.append(file_name)
                    elif mark_failure :
                        with open(os.path.join(self.local_root, task.task_work_path, 'tag_failure_download_%s' % jj), 'w') as fp: pass
                    else:
                        pass
                else:
                    file_list.append(file_name)
            if back_error:
                errors=glob(os.path.join(task.task_work_path, 'error*'))
                file_list.extend(errors)
        file_list.extend(submission.backward_common_files)
        if len(file_list) > 0:
            self._get_files(file_list)
        os.chdir(cwd)
        
    def block_checkcall(self, 
                        cmd,
                        asynchronously=False,
                        retry=0) :
        self.ssh_session.ensure_alive()
        if asynchronously:
            cmd = "nohup %s >/dev/null &" % cmd
        stdin, stdout, stderr = self.ssh_session.exec_command(('cd %s ;' % self.remote_root) + cmd)
        exit_status = stdout.channel.recv_exit_status() 
        if exit_status != 0:
            if retry < 3:
                # sleep 60 s
                dlog.warning("Get error code %d in calling %s through ssh with job: %s . message: %s" %
                        (exit_status, cmd, self.submission.submission_hash, stderr.read().decode('utf-8')))
                dlog.warning("Sleep 60 s and retry the command...")
                time.sleep(60)
                return self.block_checkcall(cmd, asynchronously=asynchronously, retry=retry+1)
            print('debug:self.remote_root, cmd', self.remote_root, cmd)
            raise RuntimeError("Get error code %d in calling %s through ssh with job: %s . message: %s" %
                               (exit_status, cmd, self.submission.submission_hash, stderr.read().decode('utf-8')))
        return stdin, stdout, stderr    

    def block_call(self, 
                   cmd) :
        self.ssh_session.ensure_alive()
        stdin, stdout, stderr = self.ssh_session.exec_command(('cd %s ;' % self.remote_root) + cmd)
        exit_status = stdout.channel.recv_exit_status() 
        return exit_status, stdin, stdout, stderr

    def clean(self) :        
        self.ssh_session.ensure_alive()
        self._rmtree(self.remote_root)

    def write_file(self, fname, write_str):
        self.ssh_session.ensure_alive()
        with self.sftp.open(os.path.join(self.remote_root, fname), 'w') as fp :
            fp.write(write_str)

    def read_file(self, fname):
        self.ssh_session.ensure_alive()
        with self.sftp.open(os.path.join(self.remote_root, fname), 'r') as fp:
            ret = fp.read().decode('utf-8')
        return ret

    def check_file_exists(self, fname):
        self.ssh_session.ensure_alive()
        try:
            self.sftp.stat(os.path.join(self.remote_root, fname)) 
            ret = True
        except IOError:
            ret = False
        return ret        
        
    def call(self, cmd):
        stdin, stdout, stderr = self.ssh_session.exec_command(cmd)
        # stdin, stdout, stderr = self.ssh.exec_command('echo $$; exec ' + cmd)
        # pid = stdout.readline().strip()
        # print(pid)
        return {'stdin':stdin, 'stdout':stdout, 'stderr':stderr}
    
    def check_finish(self, cmd_pipes):
        return cmd_pipes['stdout'].channel.exit_status_ready()


    def get_return(self, cmd_pipes):
        if not self.check_finish(cmd_pipes):
            return None, None, None
        else :
            retcode = cmd_pipes['stdout'].channel.recv_exit_status()
            return retcode, cmd_pipes['stdout'], cmd_pipes['stderr']

    def kill(self, cmd_pipes) :
        raise RuntimeError('dose not work! we do not know how to kill proc through paramiko.SSHClient')
        self.block_checkcall('kill -15 %s' % cmd_pipes['pid'])


    def _rmtree(self, remotepath, verbose = False):
        """Remove the remote path."""
        # The original implementation method removes files one by one using sftp.
        # If the latency of the remote server is high, it is very slow.
        # Thus, it's better to use system's `rm` to remove a directory, which may
        # save a lot of time.
        if verbose:
            dlog.info('removing %s' % remotepath)
        # In some supercomputers, it's very slow to remove large numbers of files
        # (e.g. directory containing trajectory) due to bad I/O performance.
        # So an asynchronously option is provided.
        self.block_checkcall('rm -rf %s' % remotepath, asynchronously=self.clean_asynchronously)

    def _put_files(self,
                   files,
                   dereference = True) :
        of = self.submission.submission_hash + '.tgz'
        # local tar
        cwd = os.getcwd()
        os.chdir(self.local_root)
        if os.path.isfile(of) :
            os.remove(of)
        with tarfile.open(of, "w:gz", dereference = dereference) as tar:
            for ii in files :
                tar.add(ii)
        os.chdir(cwd)

        try:
            self.sftp.mkdir(self.remote_root)
        except OSError: 
            pass
        # trans
        from_f = os.path.join(self.local_root, of)
        to_f = os.path.join(self.remote_root, of)
        try:
           self.sftp.put(from_f, to_f)
        except FileNotFoundError:
           raise FileNotFoundError("from %s to %s @ %s : %s Error!"%(from_f, self.ssh_session.username, self.ssh_session.hostname, to_f))
        # remote extract
        self.block_checkcall('tar xf %s' % of)
        # clean up
        os.remove(from_f)
        self.sftp.remove(to_f)

    def _get_files(self, 
                   files) :
        of = self.submission.submission_hash + '.tar.gz'
        # remote tar
        # If the number of files are large, we may get "Argument list too long" error.
        # Thus, we may run tar commands for serveral times and tar only 100 files for
        # each time.
        per_nfile = 100
        ntar = len(files) // per_nfile + 1
        if ntar <= 1:
            self.block_checkcall('tar czf %s %s' % (of, " ".join(files)))
        else:
            of_tar = self.submission.submission_hash + '.tar'
            for ii in range(ntar):
                ff = files[per_nfile * ii : per_nfile * (ii+1)]
                if ii == 0:
                    # tar cf for the first time
                    self.block_checkcall('tar cf %s %s' % (of_tar, " ".join(ff)))
                else:
                    # append using tar rf
                    # -r, --append append files to the end of an archive
                    self.block_checkcall('tar rf %s %s' % (of_tar, " ".join(ff)))
            # compress the tar file using gzip, and will get a tar.gz file
            # overwrite considering dpgen may stop and restart
            # -f, --force force overwrite of output file and compress links
            self.block_checkcall('gzip -f %s' % of_tar)
        # trans
        from_f = os.path.join(self.remote_root, of)
        to_f = os.path.join(self.local_root, of)
        if os.path.isfile(to_f) :
            os.remove(to_f)
        self.sftp.get(from_f, to_f)
        # extract
        cwd = os.getcwd()
        os.chdir(self.local_root)
        with tarfile.open(of, "r:gz") as tar:
            tar.extractall()
        os.chdir(cwd)        
        # cleanup
        os.remove(to_f)
        self.sftp.remove(from_f)



import os,sys,time,random,uuid
import psutil

from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog
from dpdispatcher.machine import Machine

shell_script_header_template="""
#!/bin/bash -l
"""

class Shell(Machine):
    def gen_script(self, job):
        shell_script = super(Shell, self).gen_script(job)
        return shell_script

    def gen_script_header(self, job):
        shell_script_header = shell_script_header_template
        return shell_script_header

    def do_submit(self, job):
        script_str = self.gen_script(job) 
        script_file_name = job.script_file_name
        job_id_name = job.job_hash + '_job_id'
        self.context.write_file(fname=script_file_name, write_str=script_str)
        proc = self.context.call('cd %s && exec bash %s' % (self.context.remote_root, script_file_name))

        job_id = int(proc.pid)
        print('shell.do_submit.job_id', job_id)
        self.context.write_file(job_id_name, str(job_id))
        return job_id

        # script_file_name = job.script_file_name
        # script_str = self.gen_script(job)
        # job_id_name = job.job_hash + '_job_id'
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        # self.context.write_file(fname=script_file_name, write_str=script_str)
        # stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'qsub', script_file_name))
        # subret = (stdout.readlines())
        # job_id = subret[0].split()[0]
        # self.context.write_file(job_id_name, job_id)        
        # return job_id


    def default_resources(self, resources) :
        pass
    
    def check_status(self, job):
        job_id = job.job_id
        print('shell.check_status.job_id', job_id)
        job_state = JobStatus.unknown
        if job_id == "" :
            return JobStatus.unsubmitted

        if_job_exists = psutil.pid_exists(pid=job_id)
        if self.check_finish_tag(job=job):
            return JobStatus.finished

        if if_job_exists:
            return JobStatus.running
        else:
            return JobStatus.terminated
        return job_state
    
    # def check_status(self, job):
    #     job_id = job.job_id
    #     uuid_names = job.job_hash
    #     cnt = 0
    #     ret, stdin, stdout, stderr = self.context.block_call("ps aux | grep %s"%uuid_names)
    #     response_list = stdout.read().decode('utf-8').split("\n")
    #     for response in response_list:
    #         if  uuid_names + ".sub" in response:
    #             return True
    #     return False

    def check_status_(self, job):
        job_id = job.job_id
        if job_id == "" :
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr\
            = self.context.block_call ("qstat -x " + job_id)
        err_str = stderr.read().decode('utf-8')
        if (ret != 0) :
            if str("qstat: Unknown Job Id") in err_str :
                if self.check_finish_tag() :
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            else :
                raise RuntimeError ("status command qstat fails to execute. erro info: %s return code %d"
                                    % (err_str, ret))
        status_line = stdout.read().decode('utf-8').split ('\n')[-2]
        status_word = status_line.split ()[-2]        
        # dlog.info (status_word)
        if status_word in ["Q","H"] :
            return JobStatus.waiting
        elif    status_word in ["R"] :
            return JobStatus.running
        elif    status_word in ["C", "E", "K", "F"] :
            if self.check_finish_tag(job) :
                return JobStatus.finished
            else :
                return JobStatus.terminated
        else :
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + '_job_tag_finished'
        print('job finished: ',job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)



import os,sys,json,glob,shutil,uuid,getpass
import unittest
from pathlib import Path

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import SSHContext, SSHSession
from .context import setUpModule

class TestSSHContext(unittest.TestCase):
    def setUp(self):
        self.tmp_local_root = 'test_context_dir/'
        self.tmp_remote_root = 'tmp_ssh_context_dir/'





from dpdispatcher.Dispatcher import Dispatcher, _split_tasks, JobRecord
from paramiko.ssh_exception import NoValidConnectionsError
import os, time
from dpdispatcher import dlog
class Entity():
    def __init__(self, ip, instance_id, job_record=None, job_handler=None):
        self.ip = ip
        self.instance_id = instance_id
        self.job_record = job_record
        self.job_handler = job_handler

class DispatcherList():
    def __init__(self, mdata_machine, mdata_resources, work_path, run_tasks, group_size, cloud_resources=None):
        self.mdata_machine = mdata_machine
        self.mdata_resources = mdata_resources
        self.task_chunks = _split_tasks(run_tasks, group_size)
        self.nchunks = len(self.task_chunks)
        self.nchunks_limit = int(self.mdata_machine.get("machine_upper_bound", self.nchunks))
        if(self.nchunks_limit > self.nchunks):
            self.nchunks_limit = self.nchunks    
        self.work_path = work_path
        self.cloud_resources = cloud_resources
        self.server_pool = []
        self.ip_pool = []
        self.dispatcher_list = list({"dispatcher": None, 
                                     "dispatcher_status": "unallocated",
                                     "entity": None} for ii in range(self.nchunks))
    # Derivate
    def init(self):
        # do something necessary
        for ii in range(self.nchunks):
            self.create(ii)

    # Base
    def run_jobs(self,
                 resources,
                 command,
                 work_path,
                 tasks,
                 group_size,
                 forward_common_files,
                 forward_task_files,
                 backward_task_files,
                 forward_task_deference = True,
                 mark_failure = False,
                 outlog = 'log',
                 errlog = 'err'):
        ratio_failure = self.mdata_resources.get("ratio_failue", 0)
        while True:
            if self.check_all_dispatchers_finished(ratio_failure):
                self.clean()
                break
            self.exception_handling(ratio_failure)
            jj = self.nchunks - 1
            for ii in range(self.nchunks):
                dispatcher_status = self.check_dispatcher_status(ii)
                if dispatcher_status == "unsubmitted":
                    dlog.info(self.dispatcher_list[ii]["entity"].ip)
                    self.dispatcher_list[ii]["entity"].job_handler = self.dispatcher_list[ii]["dispatcher"].submit_jobs(resources,
                                                                                                                        command,
                                                                                                                        work_path,
                                                                                                                        self.task_chunks[ii],
                                                                                                                        group_size,
                                                                                                                        forward_common_files,
                                                                                                                        forward_task_files,
                                                                                                                        backward_task_files,
                                                                                                                        forward_task_deference,
                                                                                                                        outlog,
                                                                                                                        errlog)
                    self.dispatcher_list[ii]["entity"].job_record = self.dispatcher_list[ii]["entity"].job_handler["job_record"]
                    self.dispatcher_list[ii]["dispatcher_status"] = "running"
                elif dispatcher_status == "finished" and self.dispatcher_list[ii]["entity"]: 
                    # no jobs in queue, delete current machine
                    # else add current machine to server_pool
                    entity = self.dispatcher_list[ii]["entity"]
                    status_list = [item["dispatcher_status"] for item in self.dispatcher_list]
                    flag = "unallocated" in status_list
                    if not flag:
                        self.delete(ii)
                        self.dispatcher_list[ii]["entity"] = None
                    else:
                        self.dispatcher_list[ii]["entity"] = None
                        self.server_pool.append(entity.instance_id)
                        self.ip_pool.append(entity.ip)
                        while(jj>=ii):
                            if(self.dispatcher_list[jj]["dispatcher_status"] == "unallocated"):
                                self.create(jj)
                                if(self.dispatcher_list[jj]["dispatcher_status"] == "unsubmitted"):
                                    dlog.info(self.dispatcher_list[jj]["entity"].ip)
                                    self.dispatcher_list[jj]["entity"].job_handler = self.dispatcher_list[jj]["dispatcher"].submit_jobs(resources,
                                                                                                                                        command,
                                                                                                                                        work_path,
                                                                                                                                        self.task_chunks[jj],
                                                                                                                                        group_size,
                                                                                                                                        forward_common_files,
                                                                                                                                        forward_task_files,
                                                                                                                                        backward_task_files,
                                                                                                                                        forward_task_deference,
                                                                                                                                        outlog,
                                                                                                                                        errlog)
                                    self.dispatcher_list[jj]["entity"].job_record = self.dispatcher_list[jj]["entity"].job_handler["job_record"]
                                    self.dispatcher_list[jj]["dispatcher_status"] = "running"
                                break
                            jj -=1
                elif dispatcher_status == "running":
                    pass
                elif dispatcher_status == "unallocated":
                    # if len(server_pool) > 0: make_dispatcher
                    # else: pass
                    self.create(ii)
                    if self.dispatcher_list[ii]["dispatcher_status"] == "unsubmitted":
                        dlog.info(self.dispatcher_list[ii]["entity"].ip)
                        self.dispatcher_list[ii]["entity"].job_handler = self.dispatcher_list[ii]["dispatcher"].submit_jobs(resources,
                                                                                                                            command,
                                                                                                                            work_path,
                                                                                                                            self.task_chunks[ii],
                                                                                                                            group_size,
                                                                                                                            forward_common_files,
                                                                                                                            forward_task_files,
                                                                                                                            backward_task_files,
                                                                                                                            forward_task_deference,
                                                                                                                            outlog,
                                                                                                                            errlog)
                        self.dispatcher_list[ii]["entity"].job_record = self.dispatcher_list[ii]["entity"].job_handler["job_record"]
                        self.dispatcher_list[ii]["dispatcher_status"] = "running"
                elif dispatcher_status == "terminated":
                    pass
            self.update()
            time.sleep(10)    

    # Derivate
    def create(self, ii):
        '''case1: use existed machine(finished) to make_dispatcher
           case2: create one machine, then make_dispatcher, change status from unallocated to unsubmitted'''
        pass
    
    # Derivate
    def delete(self, ii):
        '''delete one machine
            if entity is none, means this machine is used by another dispatcher, shouldn't be deleted'''
        pass

    # Derivate, delete config like templates, etc.
    def clean(self):
        pass

    # Derivate
    def update():
        pass

    # Base
    def check_all_dispatchers_finished(self, ratio_failure):    
        status_list = [item["dispatcher_status"] for item in self.dispatcher_list]
        finished_num = status_list.count("finished")
        if finished_num / self.nchunks < (1 - ratio_failure): return False
        else: return True

    # Base
    def exception_handling(self, ratio_failure):
        status_list = [item["dispatcher_status"] for item in self.dispatcher_list]
        terminated_num = status_list.count("terminated")
        if terminated_num / self.nchunks > ratio_failure:
            # self.dispatcher_list = [lambda item["dispatcher_status"]: "finished" for item in self.dispatcher_list if item["dispatcher_status"] == "terminated"]
            for ii in range(self.nchunks):
                if self.dispatcher_list[ii]["dispatcher_status"] == "terminated":
                    self.dispatcher_list[ii]["dispatcher_status"] = "unallocated"
    # Base
    def make_dispatcher(self, ii):
        entity = self.dispatcher_list[ii]["entity"]
        profile = self.mdata_machine.copy()
        profile['hostname'] = entity.ip
        profile['instance_id'] = entity.instance_id
        count = 0
        flag = 0
        while count < 3:
            try:
                self.dispatcher_list[ii]["dispatcher"] = Dispatcher(profile, context_type='ssh', batch_type='shell', job_record='jr.%.06d.json' % ii)
                self.dispatcher_list[ii]["dispatcher_status"] = "unsubmitted"
                flag = 1
                break
            except:
                count += 1
                time.sleep(60)
        if not flag:
            # give up this machine, wait other machine in sever_pool.
            # this machine will be append into server_pool next time when update apg_instances.
            self.dispatcher_list[ii]["entity"] = None


    # Base
    def check_dispatcher_status(self, ii, allow_failue=False):
        '''catch running dispatcher exception
           if no exception occured, check finished'''
        if self.dispatcher_list[ii]["dispatcher_status"] == "running":
            status = self.catch_dispatcher_exception(ii)
            if status == 0:
                # param clean: delete remote work_dir or not.
                clean = self.mdata_resources.get("clean", False)
                try:
                    # avoid raising ssh exception in download proceess
                    finished = self.dispatcher_list[ii]["dispatcher"].all_finished(self.dispatcher_list[ii]["entity"].job_handler, allow_failue, clean)
                    if finished:
                        self.dispatcher_list[ii]["dispatcher_status"] = "finished"
                except:
                    pass                    
            elif status == 1:
                # self.dispatcher_list[ii]["dispatcher_status"] = "terminated"
                pass
            elif status == 2:
                self.dispatcher_list[ii]["dispatcher"] = None
                self.dispatcher_list[ii]["dispatcher_status"] = "terminated"
                self.dispatcher_list[ii]["entity"] = None
                os.remove(os.path.join(self.work_path, "jr.%.06d.json" % ii))
        return self.dispatcher_list[ii]["dispatcher_status"]

    # Derivate
    def catch_dispatcher_exception(self, ii):
        '''everything is okay: return 0
           ssh not active    : return 1
           machine callback  : return 2'''
        pass









import os, glob
from posixpath import realpath
from zipfile import ZipFile

def zip_file_list(root_path, zip_filename, file_list=[]):
    out_zip_file = os.path.join(root_path, zip_filename)
    zip_obj = ZipFile(out_zip_file, 'w')
    for f in file_list:
        matched_files = os.path.join(root_path, f)
        for ii in glob.glob(matched_files):
            realpath = os.path.realpath(ii)
            arcname = os.path.relpath(ii, start=root_path)
        zip_obj.write(realpath, arcname)
    zip_obj.close()
    return out_zip_file

def zip_files(root_path, out_file, selected=[]):
    obj = ZipFile(out_file, "w")
    # change /xxx/ to /xxx or xxx to /xxx and pop ''
    for i in range(len(selected)):
        if not selected[i]:
            selected.pop(i)
            continue

        selected[i] = selected[i].strip()
        if selected[i].endswith('/'):
            selected[i] = selected[i][:-1]
        if not selected[i].startswith('/'):
            selected[i] = '/{}'.format(selected[i])

    for root, dirs, files in os.walk(root_path):
        for item in files:
            filename = os.path.join(root, item)
            arcname = filename.replace(root_path,'')
            if not is_selected(arcname, selected):
                continue

            obj.write(filename, arcname)
    if not obj.filelist:
        return

    obj.close()


def is_selected(arcname, selected):
    if not selected:
        return True

    arcdir = os.path.dirname(arcname)
    for s in selected:
        if arcname == s:
            return True

        if arcdir == s:
            return True

        if arcname.startswith(s + '/'):
            return True

    return False


def unzip_file(zip_file, out_dir='./'):
    obj = ZipFile(zip_file, "r")
    for item in obj.namelist():
        obj.extract(item, out_dir)



import os, getpass, time
from abc import ABC

from dpdispatcher.machine import Machine
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher.submission import Resources


lsf_script_template = """\
{lsf_script_header}
{lsf_script_env}
{lsf_script_command}
{lsf_script_end}
"""

lsf_script_header_template = """\
#!/bin/bash -l
#BSUB -e %J.err
#BSUB -o %J.out
{lsf_nodes_line}
{lsf_ptile_line}
{lsf_partition_line}
{lsf_walltime_line}
{lsf_number_gpu_line}
"""

lsf_script_env_template = """
{prepend_text}
export REMOTE_ROOT={remote_root}
test $? -ne 0 && exit 1
"""

lsf_script_command_template = """
cd $REMOTE_ROOT
cd {task_work_path}
test $? -ne 0 && exit 1
if [ ! -f {task_tag_finished} ] ;then
  {command_env} {command}  1>> {outlog} 2>> {errlog} 
  if test $? -ne 0; then touch {task_tag_finished}; fi
  touch {task_tag_finished}
fi 
"""

lsf_script_end_template = """

cd $REMOTE_ROOT
test $? -ne 0 && exit 1

wait

{append_text}
touch {job_tag_finished}
"""

lsf_script_wait = """
wait
"""

default_lsf_bsub_dict = {
    'm': "8G"
}


# class LSFResources(Resources):
#     """
#     LSF resources
#     """
#     def __init__(
#             self,
#             number_node,
#             cpu_per_node,
#             gpu_per_node,
#             queue_name,
#             walltime="120:00:00",
#             prepend_text="",
#             append_text="",
#             gpu_usage=True,
#             gpu_new_syntax=True,
#             lsf_bsub_dict=None,
#             group_size=1
#     ):
#         """define LSF resources
#
#         Parameters
#         ----------
#         number_node: nodes to be used
#         cpu_per_node: CPU cores used on each node
#         gpu_per_node: GPU
#         queue_name: the name of queue
#         walltime: max time of task
#         prepend_text: prepend scripts, code executed before the task run
#         append_text: append scripts, code executed after the task run
#         gpu_usage: choose if GPU line is used
#         lsf_bsub_dict: other bsub parameters.
#         group_size: tasks contained by each group
#         """
#         super().__init__(number_node, cpu_per_node, gpu_per_node, queue_name, group_size)
#         if lsf_bsub_dict is None:
#             lsf_bsub_dict = {}
#         self.walltime = walltime
#         self.gpu_new_syntax = gpu_new_syntax
#         self.gpu_usage = gpu_usage
#         self.prepend_text = prepend_text
#         self.append_text = append_text
#         self.lsf_bsub_dict = lsf_bsub_dict


class LSF(Machine):
    """
    LSF batch
    """
    def gen_script(self, job):
        # if type(job.resources) is LSFResources:
        #     resources = job.resources
        #     lsf_bsub_dict = job.resources.lsf_bsub_dict
        #     if lsf_bsub_dict is None:
        #         lsf_bsub_dict = {}
        # elif type(job.resources) is Resources:
        #     resources = LSFResources(**job.resources.__dict__)
        #     lsf_bsub_dict = {}
        # else:
        #     raise RuntimeError('type job.resource error')

        resources = job.resources
        lsf_bsub_dict = resources.extra_specification.copy()

        # headers
        script_header_dict = {
            'lsf_nodes_line': "#BSUB -n {number_cores}".format(
                number_cores=resources.number_node * resources.cpu_per_node),
            'lsf_ptile_line': "#BSUB -R 'span[ptile={cpu_per_node}]'".format(
                cpu_per_node=resources.cpu_per_node),
            'lsf_partition_line': "#BSUB -q {queue_name}".format(
                queue_name=resources.queue_name),
            'lsf_walltime_line': "#BSUB -W {walltime}".format(
                walltime=resources.kwargs.get('walltime', '12:00'))
        }
        gpu_usage_flag = resources.kwargs.get('gpu_usage', False)
        gpu_new_syntax_flag = resources.kwargs.get('gpu_new_syntax', False)
        if gpu_usage_flag is True:
            if gpu_new_syntax_flag is True:
                script_header_dict['lsf_number_gpu_line'] = "#BSUB -gpu 'num={gpu_per_node}:mode=shared:" \
                                                            "j_exclusive=yes'".format(
                    gpu_per_node=resources.gpu_per_node)
            else:
                script_header_dict['lsf_number_gpu_line'] = '#BSUB -R "select[ngpus >0] rusage[' \
                                                            'ngpus_excl_p={gpu_per_node}]"'.format(
                    gpu_per_node=resources.gpu_per_node)
        else:
            script_header_dict['lsf_number_gpu_line'] = ""
        lsf_script_header = lsf_script_header_template.format(**script_header_dict)

        for k, v in lsf_bsub_dict.items():
            line = "#BSUB -{key} {value}\n".format(key=k, value=str(v))
            lsf_script_header += line

        # envs
        script_env_dict = {
            'prepend_text': resources.kwargs.get('prepend_text', ""),
            'remote_root': self.context.remote_root
        }
        lsf_script_env = lsf_script_env_template.format(**script_env_dict)

        # commands
        lsf_script_command = ""

        for task in job.job_task_list:
            command_env = ""
            lsf_script_command += self.get_script_wait(resources=resources, task=task)
            command_env += self.get_command_env_cuda_devices(resources=resources, task=task)

            task_tag_finished = task.task_hash + '_task_tag_finished'

            temp_lsf_script_command = lsf_script_command_template.format(
                command_env=command_env,
                task_work_path=task.task_work_path,
                command=task.command,
                task_tag_finished=task_tag_finished,
                outlog=task.outlog, errlog=task.errlog
            )

            lsf_script_command += temp_lsf_script_command

        # end
        job_tag_finished = job.job_hash + '_job_tag_finished'
        lsf_script_end = lsf_script_end_template.format(
            append_text=resources.kwargs.get('append_text', ""),
            job_tag_finished=job_tag_finished
        )

        # join the whole script
        lsf_script = lsf_script_template.format(
            lsf_script_header=lsf_script_header,
            lsf_script_env=lsf_script_env,
            lsf_script_command=lsf_script_command,
            lsf_script_end=lsf_script_end)
        return lsf_script

    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + '_job_id'
        self.context.write_file(fname=script_file_name, write_str=script_str)
        stdin, stdout, stderr = self.context.block_checkcall(
            'cd %s && %s %s' % (self.context.remote_root, 'bsub < ', script_file_name)
        )
        subret = (stdout.readlines())
        job_id = subret[0].split()[1][1:-1]
        self.context.write_file(job_id_name, job_id)
        return job_id

    # TODO: add default resources
    def default_resources(self, resources):
        pass

    def check_status(self, job):
        try:
            job_id = job.job_id
        except AttributeError:
            return JobStatus.terminated
        if job_id == "":
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr \
            = self.context.block_call("bjobs " + job_id)
        err_str = stderr.read().decode('utf-8')
        if ("Job <%s> is not found" % job_id) in err_str:
            if self.check_finish_tag(job):
                return JobStatus.finished
            else:
                return JobStatus.terminated
        elif ret != 0:
            raise RuntimeError("status command bjobs fails to execute.\n error info: %s \nreturn code %d\n"
                               % (err_str, ret))
        status_out = stdout.read().decode('utf-8').split('\n')
        if len(status_out) < 2:
            return JobStatus.unknown
        else:
            status_line = status_out[1]
            status_word = status_line.split()[2]

        # ref: https://www.ibm.com/support/knowledgecenter/en/SSETD4_9.1.2/lsf_command_ref/bjobs.1.html
        if status_word in ["PEND", "WAIT", "PSUSP"]:
            return JobStatus.waiting
        elif status_word in ["RUN", "USUSP"]:
            return JobStatus.running
        elif status_word in ["DONE", "EXIT"]:
            if self.check_finish_tag(job):
                return JobStatus.finished
            else:
                return JobStatus.terminated
        else:
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + '_job_tag_finished'
        print('job finished: ', job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)



import os,sys,time,random,uuid

from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog
from dpdispatcher.machine import Machine


pbs_script_header_template="""
#!/bin/bash -l
{select_node_line}
#PBS -j oe
{queue_name_line}
"""

class PBS(Machine):
    def gen_script(self, job):
        pbs_script = super(PBS, self).gen_script(job)
        return pbs_script

    def gen_script_header(self, job):
        resources = job.resources
        pbs_script_header_dict= {}
        pbs_script_header_dict['select_node_line']="#PBS -l select={number_node}:ncpus={cpu_per_node}:ngpus={gpu_per_node}".format(
            number_node=resources.number_node, cpu_per_node=resources.cpu_per_node, gpu_per_node=resources.gpu_per_node
        )
        pbs_script_header_dict['queue_name_line']="#PBS -q {queue_name}".format(queue_name=resources.queue_name)
        pbs_script_header = pbs_script_header_template.format(**pbs_script_header_dict)
        return pbs_script_header

    def do_submit(self, job):
        script_file_name = job.script_file_name
        script_str = self.gen_script(job)
        job_id_name = job.job_hash + '_job_id'
        # script_str = self.sub_script(job_dirs, cmd, args=args, resources=resources, outlog=outlog, errlog=errlog)
        self.context.write_file(fname=script_file_name, write_str=script_str)
        # self.context.write_file(fname=os.path.join(self.context.submission.work_base, script_file_name), write_str=script_str)
        # script_file_dir = os.path.join(self.context.submission.work_base)
        script_file_dir = self.context.remote_root
        # stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (self.context.remote_root, 'qsub', script_file_name))
        stdin, stdout, stderr = self.context.block_checkcall('cd %s && %s %s' % (script_file_dir, 'qsub', script_file_name))
        subret = (stdout.readlines())
        job_id = subret[0].split()[0]
        self.context.write_file(job_id_name, job_id)        
        return job_id


    def default_resources(self, resources) :
        pass
    
    def check_status(self, job):
        job_id = job.job_id
        if job_id == "" :
            return JobStatus.unsubmitted
        ret, stdin, stdout, stderr\
            = self.context.block_call ("qstat -x " + job_id)
        err_str = stderr.read().decode('utf-8')
        if (ret != 0) :
            if str("qstat: Unknown Job Id") in err_str or str("Job has finished") in err_str:
                if self.check_finish_tag(job=job) :
                    return JobStatus.finished
                else :
                    return JobStatus.terminated
            else :
                raise RuntimeError ("status command qstat fails to execute. erro info: %s return code %d"
                                    % (err_str, ret))
        status_line = stdout.read().decode('utf-8').split ('\n')[-2]
        status_word = status_line.split ()[-2]        
        # dlog.info (status_word)
        if status_word in ["Q","H"] :
            return JobStatus.waiting
        elif    status_word in ["R"] :
            return JobStatus.running
        elif    status_word in ["C", "E", "K", "F"] :
            if self.check_finish_tag(job) :
                return JobStatus.finished
            else :
                return JobStatus.terminated
        else :
            return JobStatus.unknown

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + '_job_tag_finished'
        print('check if job finished: ',job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)






from aliyunsdkecs.request.v20140526.DescribeInstancesRequest import DescribeInstancesRequest
from aliyunsdkcore.client import AcsClient
import aliyunsdkcore.request
aliyunsdkcore.request.set_default_protocol_type("https")
from aliyunsdkcore.acs_exception.exceptions import ClientException
from aliyunsdkcore.acs_exception.exceptions import ServerException
from aliyunsdkecs.request.v20140526.RunInstancesRequest import RunInstancesRequest
from aliyunsdkecs.request.v20140526.DeleteInstancesRequest import DeleteInstancesRequest
from aliyunsdkecs.request.v20140526.DescribeAutoProvisioningGroupInstancesRequest import DescribeAutoProvisioningGroupInstancesRequest
from aliyunsdkecs.request.v20140526.CreateAutoProvisioningGroupRequest import CreateAutoProvisioningGroupRequest
from aliyunsdkecs.request.v20140526.DeleteAutoProvisioningGroupRequest import DeleteAutoProvisioningGroupRequest
from aliyunsdkecs.request.v20140526.ModifyAutoProvisioningGroupRequest import ModifyAutoProvisioningGroupRequest
from aliyunsdkecs.request.v20140526.DeleteLaunchTemplateRequest import DeleteLaunchTemplateRequest
from aliyunsdkvpc.request.v20160428.DescribeVpcsRequest import DescribeVpcsRequest
from aliyunsdkecs.request.v20140526.DescribeLaunchTemplatesRequest import DescribeLaunchTemplatesRequest
from aliyunsdkecs.request.v20140526.CreateLaunchTemplateRequest import CreateLaunchTemplateRequest
from aliyunsdkecs.request.v20140526.DescribeImagesRequest import DescribeImagesRequest
from aliyunsdkecs.request.v20140526.DescribeSecurityGroupsRequest import DescribeSecurityGroupsRequest
from aliyunsdkvpc.request.v20160428.DescribeVSwitchesRequest import DescribeVSwitchesRequest
import time, json, os, glob, string, random, sys
from dpdispatcher.Dispatcher import Dispatcher, _split_tasks, JobRecord
from dpdispatcher.SSHContext import SSHSession
from dpdispatcher.DispatcherList import DispatcherList, Entity
from os.path import join
from dpdispatcher  import dlog
from hashlib import sha1

# cloud_resources = {"AccessKey_ID":"",
#                    "AccessKey_Secret":"",
#                    "regionID": "cn-shenzhen",
#                    "img_name": "kit",
#                    "machine_type_price": [
#                        {"machine_type": "ecs.gn6v-c8g1.2xlarge", "price_limit": 20.00, "numb": 1, "priority": 0},
#                        {"machine_type": "ecs.gn5-c4g1.xlarge",   "price_limit": 20.00, "numb": 1, "priority": 1}
#                    ],
#                    "instance_name": "CH4_test_username", 
#                    "pay_strategy": "spot"
#                    "apg_id": apg_id, 
#                    "template_id": template_id, 
#                    "vsw_id": vsw_id, 
#                    "region_id": region_id,
#                    "client": client}

def manual_create(stage, num):
    '''running this function in your project root path, which contains machine-ali.json.
       please ensure your machine name is machine-ali.json 
       This will create a subdir named manual, which includes apg_id.json'''
    root_path = os.getcwd()
    fp = open("machine-ali.json")
    data = json.load(fp)
    if not os.path.exists("manual"):
        os.mkdir("manual")
    os.chdir("manual")
    mdata_machine = data[stage][0]["machine"]
    mdata_resources = data[stage][0]["resources"]
    cloud_resources = mdata_machine["cloud_resources"]
    ali = ALI(mdata_machine, mdata_resources, "work_path", [1], 1, cloud_resources)
    img_id = ali.get_image_id(ali.cloud_resources["img_name"])
    sg_id, vpc_id = ali.get_sg_vpc_id()
    ali.cloud_resources["template_id"] = ali.create_template(img_id, sg_id, vpc_id)
    ali.cloud_resources["vsw_id"] = ali.get_vsw_id(vpc_id)
    ali.nchunks_limit = num 
    ali.cloud_resources["apg_id"] = ali.create_apg()
    time.sleep(90)
    instance_list = ali.describe_apg_instances()
    ip_list = ali.get_ip(instance_list)
    print(instance_list)
    print(ip_list)

def manual_delete(stage):
    '''running this function in your project root path, which contains machine-ali.json. '''
    if os.path.exists("manual"):
        fp = open("machine-ali.json")
        data = json.load(fp)
        mdata_machine = data[stage][0]["machine"]
        mdata_resources = data[stage][0]["resources"]
        cloud_resources = mdata_machine["cloud_resources"]
        ali = ALI(mdata_machine, mdata_resources, "work_path", [1], 1, cloud_resources)
        os.chdir("manual")
        fp = open("apg_id.json")
        data = json.load(fp)
        ali.cloud_resources["apg_id"] = data["apg_id"]
        ali.delete_apg()
        os.remove("apg_id.json")
        print("delete successfully!")

class ALI(DispatcherList):
    def __init__(self, mdata_machine, mdata_resources, work_path, run_tasks, group_size, cloud_resources=None):
        super().__init__(mdata_machine, mdata_resources, work_path, run_tasks, group_size, cloud_resources)
        self.client = AcsClient(cloud_resources["AccessKey_ID"], cloud_resources["AccessKey_Secret"], cloud_resources["regionID"])

    def init(self):
        self.prepare()
        for ii in range(self.nchunks):
            self.create(ii)

    def create(self, ii):
        if self.dispatcher_list[ii]["dispatcher_status"] == "unallocated" and len(self.ip_pool) > 0:
            self.dispatcher_list[ii]["entity"] = Entity(self.ip_pool.pop(0), self.server_pool.pop(0))
            self.make_dispatcher(ii)

    # Derivate
    def delete(self, ii):
        '''delete one machine'''
        request = DeleteInstancesRequest()
        request.set_accept_format('json')
        request.set_InstanceIds([self.dispatcher_list[ii]["entity"].instance_id])
        request.set_Force(True)
        count = 0
        flag = 0
        while count < 10:
            try:
                response = self.client.do_action_with_exception(request)
                flag = 1
                break
            except ServerException as e:
                time.sleep(10)
                count += 1
                
        if flag:
            status_list = [item["dispatcher_status"] for item in self.dispatcher_list]
            running_num = status_list.count("running")
            running_num += status_list.count("unsubmitted")
            self.change_apg_capasity(running_num)
        else:
            dlog.info("delete failed, exit")
            sys.exit()

    def update(self):
        self.server_pool = self.get_server_pool()
        self.ip_pool = self.get_ip(self.server_pool)

    # Derivate
    def catch_dispatcher_exception(self, ii):
        '''everything is okay: return 0
           ssh not active    : return 1
           machine callback  : return 2'''
        if self.check_spot_callback(self.dispatcher_list[ii]["entity"].instance_id): 
            dlog.info("machine %s callback, ip: %s" % (self.dispatcher_list[ii]["entity"].instance_id, self.dispatcher_list[ii]["entity"].ip))
            return 2
        elif not self.dispatcher_list[ii]["dispatcher"].session._check_alive():
            try:
                self.dispatcher_list[ii]["dispatcher"].session.ensure_alive()
                return 0
            except RuntimeError:
                return 1
        else: return 0
        
    def get_server_pool(self):
        running_server = self.describe_apg_instances()
        allocated_server = []
        for ii in range(self.nchunks):
            if self.dispatcher_list[ii]["dispatcher_status"] == "running" or self.dispatcher_list[ii]["dispatcher_status"] == "unsubmitted":
                allocated_server.append(self.dispatcher_list[ii]["entity"].instance_id)
        return list(set(running_server) - set(allocated_server))

    def clean(self):
        self.delete_apg()
        self.delete_template()
        os.remove("apg_id.json")

    def prepare(self):
        restart = False
        if os.path.exists('apg_id.json'):
            with open('apg_id.json') as fp:
                apg = json.load(fp)
                self.cloud_resources["apg_id"] = apg["apg_id"]
            task_chunks_str = ['+'.join(ii) for ii in self.task_chunks]
            task_hashes = [sha1(ii.encode('utf-8')).hexdigest() for ii in task_chunks_str]
            for ii in range(self.nchunks):
                fn = 'jr.%.06d.json' % ii
                if os.path.exists(os.path.join(os.path.abspath(self.work_path), fn)):
                    cur_hash = task_hashes[ii]
                    job_record = JobRecord(self.work_path, self.task_chunks[ii], fn)
                    if not job_record.check_finished(cur_hash): 
                        if not self.check_spot_callback(job_record.record[cur_hash]['context']['instance_id']):
                            self.dispatcher_list[ii]["entity"] = Entity(job_record.record[cur_hash]['context']['ip'], job_record.record[cur_hash]['context']['instance_id'], job_record)
                            self.make_dispatcher(ii)
                            self.dispatcher_list[ii]["dispatcher_status"] = "unsubmitted"
                        else:
                            os.remove(os.path.join(os.path.abspath(self.work_path), fn))
                    else:
                        self.dispatcher_list[ii]["dispatcher_status"] = "finished"
            self.server_pool = self.get_server_pool()
            self.ip_pool = self.get_ip(self.server_pool)
            restart = True
        img_id = self.get_image_id(self.cloud_resources["img_name"])
        sg_id, vpc_id = self.get_sg_vpc_id()
        self.cloud_resources["template_id"] = self.create_template(img_id, sg_id, vpc_id)
        self.cloud_resources["vsw_id"] = self.get_vsw_id(vpc_id)
        if not restart:
            dlog.info("begin to create apg")
            self.cloud_resources["apg_id"] = self.create_apg()
            time.sleep(120)
            self.server_pool = self.get_server_pool()
            self.ip_pool = self.get_ip(self.server_pool)
        else: dlog.info("restart dispatcher")
        
    def delete_apg(self):
        request = DeleteAutoProvisioningGroupRequest()
        request.set_accept_format('json')
        request.set_AutoProvisioningGroupId(self.cloud_resources["apg_id"])
        request.set_TerminateInstances(True)
        count = 0
        flag = 0
        while count < 10:
            try:
                response = self.client.do_action_with_exception(request)
                flag = 1
                break
            except ServerException as e:
                time.sleep(10)
                count += 1
        if not flag:
            dlog.info("delete apg failed, exit")
            sys.exit()

        
    def create_apg(self):
        request = CreateAutoProvisioningGroupRequest()
        request.set_accept_format('json')
        request.set_TotalTargetCapacity(str(self.nchunks_limit))
        request.set_LaunchTemplateId(self.cloud_resources["template_id"])
        request.set_AutoProvisioningGroupName(self.cloud_resources["instance_name"] + ''.join(random.choice(string.ascii_uppercase) for _ in range(20)))
        request.set_AutoProvisioningGroupType("maintain")
        request.set_SpotAllocationStrategy("lowest-price")
        request.set_SpotInstanceInterruptionBehavior("terminate")
        request.set_SpotInstancePoolsToUseCount(1)
        request.set_ExcessCapacityTerminationPolicy("termination")
        request.set_TerminateInstances(True)
        request.set_PayAsYouGoTargetCapacity("0")
        request.set_SpotTargetCapacity(str(self.nchunks_limit))
        config = self.generate_config()
        request.set_LaunchTemplateConfigs(config)

        try:
            response = self.client.do_action_with_exception(request)
            response = json.loads(response)
            with open('apg_id.json', 'w') as fp:
                json.dump({'apg_id': response["AutoProvisioningGroupId"]}, fp, indent=4)
            return response["AutoProvisioningGroupId"]
        except ServerException as e:
            dlog.info("create apg failed, err msg: %s" % e)
            sys.exit()
        except ClientException as e:
            dlog.info("create apg failed, err msg: %s" % e)
            sys.exit()

    def describe_apg_instances(self):
        request = DescribeAutoProvisioningGroupInstancesRequest()
        request.set_accept_format('json')
        request.set_AutoProvisioningGroupId(self.cloud_resources["apg_id"])
        request.set_PageSize(100)
        iteration = self.nchunks // 100
        instance_list = []
        for i in range(iteration + 1):
            request.set_PageNumber(i+1)
            count = 0
            flag = 0
            err_msg = 0
            while count < 10:
                try:
                    response = self.client.do_action_with_exception(request)
                    response = json.loads(response)
                    for ins in response["Instances"]["Instance"]:
                        instance_list.append(ins["InstanceId"])
                    flag = 1
                    break
                except ServerException as e:
                    # dlog.info(e)
                    err_msg = e
                    count += 1
                except ClientException as e:
                    # dlog.info(e)
                    err_msg = e
                    count += 1
            if not flag:
                dlog.info("describe_apg_instances failed, err msg: %s" %err_msg)
                sys.exit()
        return instance_list
        
    def generate_config(self):
        machine_config = self.cloud_resources["machine_type_price"]
        config = []
        for conf in machine_config:
            for vsw in self.cloud_resources["vsw_id"]:
                tmp = {
                    "InstanceType": conf["machine_type"],
                    "MaxPrice": str(conf["price_limit"] * conf["numb"]),
                    "VSwitchId": vsw,
                    "WeightedCapacity": "1",
                    "Priority": str(conf["priority"])
                }
                config.append(tmp)
        return config

    def create_template(self, image_id, sg_id, vpc_id):
        request = CreateLaunchTemplateRequest()
        request.set_accept_format('json')
        request.set_LaunchTemplateName(''.join(random.choice(string.ascii_uppercase) for _ in range(20)))
        request.set_ImageId(image_id)
        request.set_ImageOwnerAlias("self")
        request.set_PasswordInherit(True)
        if "address" in self.cloud_resources and self.cloud_resources['address'] == "public":
            request.set_InternetMaxBandwidthIn(100)
            request.set_InternetMaxBandwidthOut(100)
        request.set_InstanceType("ecs.c6.large")
        request.set_InstanceName(self.cloud_resources["instance_name"])
        request.set_SecurityGroupId(sg_id)
        request.set_VpcId(vpc_id)
        request.set_SystemDiskCategory("cloud_efficiency")
        request.set_SystemDiskSize(70)
        request.set_IoOptimized("optimized")
        request.set_InstanceChargeType("PostPaid")
        request.set_NetworkType("vpc")
        request.set_SpotStrategy("SpotWithPriceLimit")
        request.set_SpotPriceLimit(100)
        try:
            response = self.client.do_action_with_exception(request)
            response = json.loads(response)
            return response["LaunchTemplateId"]
        except ServerException as e:
            dlog.info(e)
            sys.exit()
        except ClientException as e:
            dlog.info(e)
            sys.exit()
            
    def delete_template(self):
        request = DeleteLaunchTemplateRequest()
        request.set_accept_format('json')
        count = 0
        flag = 0
        while count < 10:
            try:
                request.set_LaunchTemplateId(self.cloud_resources["template_id"])
                response = self.client.do_action_with_exception(request)
                flag = 1
                break
            except:
                count += 1
        # count = 10 and still failed, continue

    def get_image_id(self, img_name):
        request = DescribeImagesRequest()
        request.set_accept_format('json')
        request.set_ImageOwnerAlias("self")
        request.set_PageSize(20)
        response = self.client.do_action_with_exception(request)
        response = json.loads(response)
        totalcount = response["TotalCount"]

        iteration = totalcount // 20
        if iteration * 20 < totalcount:
            iteration += 1

        for ii in range(1, iteration+1):
            count = 0
            flag = 0
            request.set_PageNumber(ii)
            while count < 10:
                try:
                    response = self.client.do_action_with_exception(request)
                    response = json.loads(response)
                    for img in response["Images"]["Image"]:
                        if img["ImageName"] == img_name:
                            return img["ImageId"]
                    flag = 1
                    break
                except:
                    count += 1
                    time.sleep(10)
        if not flag:
            dlog.info("get image failed, exit")
            sys.exit()

    def get_sg_vpc_id(self):
        request = DescribeSecurityGroupsRequest()
        request.set_accept_format('json')
        response = self.client.do_action_with_exception(request)
        response = json.loads(response)
        for sg in response["SecurityGroups"]["SecurityGroup"]:
            if sg["SecurityGroupName"] == "sg":
                return sg["SecurityGroupId"], sg["VpcId"]

    def get_vsw_id(self, vpc_id):
        request = DescribeVpcsRequest()
        request.set_accept_format('json')
        request.set_VpcId(vpc_id)
        response = self.client.do_action_with_exception(request)
        response = json.loads(response)
        for vpc in response["Vpcs"]["Vpc"]:
            if vpc["VpcId"] == vpc_id:
                vswitchids = vpc["VSwitchIds"]["VSwitchId"]
                break
        vswitchid_option = []
        if "zone" in self.cloud_resources and self.cloud_resources['zone']:
            for zone in self.cloud_resources['zone']:
                for vswitchid in vswitchids:
                    request = DescribeVSwitchesRequest()
                    request.set_accept_format('json')
                    request.set_VSwitchId(vswitchid)
                    zoneid = self.cloud_resources['regionID']+"-"+zone
                    request.set_ZoneId(zoneid)
                    response = self.client.do_action_with_exception(request)
                    response = json.loads(response)
                    if(response["TotalCount"] == 1):
                        vswitchid_option.append(vswitchid)
                        continue
        if(vswitchid_option):
            return vswitchid_option
        else:
            return  vswitchids

    def change_apg_capasity(self, capasity):
        request = ModifyAutoProvisioningGroupRequest()
        request.set_accept_format('json')
        request.set_AutoProvisioningGroupId(self.cloud_resources["apg_id"])
        request.set_TotalTargetCapacity(str(capasity))
        request.set_SpotTargetCapacity(str(capasity))
        request.set_PayAsYouGoTargetCapacity("0")
        count = 0
        flag = 0
        while count < 10:
            try:
                response = self.client.do_action_with_exception(request)
                flag = 1
                break
            except:
                count += 1
                time.sleep(10)
        if not flag:
            dlog.info("change_apg_capasity failed, exit")
            sys.exit()

    def check_spot_callback(self, instance_id):
        request = DescribeInstancesRequest()
        request.set_accept_format('json')
        request.set_InstanceIds([instance_id])
        status = False
        count = 0
        while count < 10:
            try:
                response = self.client.do_action_with_exception(request)
                response = json.loads(response)
                if len(response["Instances"]["Instance"]) == 1 and "Recycling" in response["Instances"]["Instance"][0]["OperationLocks"]["LockReason"]:
                    status = True
                if instance_id not in self.describe_apg_instances():
                    status = True
                break
            except ServerException as e:
                # dlog.info(e)
                count += 1
                time.sleep(10)
            except ClientException as e:
                # dlog.info(e)
                count += 1
                time.sleep(10)
        return status

    def get_ip(self, instance_list):
        request = DescribeInstancesRequest()
        request.set_accept_format('json')
        ip_list = []
        if len(instance_list) == 0: return ip_list
        try:
            if len(instance_list) <= 10:
                for i in range(len(instance_list)):
                    request.set_InstanceIds([instance_list[i]])
                    response = self.client.do_action_with_exception(request)
                    response = json.loads(response)
                    if "address" in self.cloud_resources and self.cloud_resources['address'] == "public":
                        ip_list.append(response["Instances"]["Instance"][0]["PublicIpAddress"]["IpAddress"][0])
                    else:
                        ip_list.append(response["Instances"]["Instance"][0]["VpcAttributes"]["PrivateIpAddress"]['IpAddress'][0])
                    # ip_list.append(response["Instances"]["Instance"][0]["PublicIpAddress"]["IpAddress"][0])
            else:
                iteration = len(instance_list) // 10
                for i in range(iteration):
                    for j in range(10):
                        request.set_InstanceIds([instance_list[i*10+j]])
                        response = self.client.do_action_with_exception(request)
                        response = json.loads(response)
                        if "address" in self.cloud_resources and self.cloud_resources['address'] == "public":
                            ip_list.append(response["Instances"]["Instance"][0]["PublicIpAddress"]["IpAddress"][0])
                        else:
                            ip_list.append(response["Instances"]["Instance"][0]["VpcAttributes"]["PrivateIpAddress"]['IpAddress'][0])
                if len(instance_list) - iteration * 10 != 0:
                    for j in range(len(instance_list) - iteration * 10):
                        request.set_InstanceIds([instance_list[iteration*10+j]])
                        response = self.client.do_action_with_exception(request)
                        response = json.loads(response)
                        if "address" in self.cloud_resources and self.cloud_resources['address'] == "public":
                            ip_list.append(response["Instances"]["Instance"][0]["PublicIpAddress"]["IpAddress"][0])
                        else:
                            ip_list.append(response["Instances"]["Instance"][0]["VpcAttributes"]["PrivateIpAddress"]['IpAddress'][0])
            return ip_list
        except: return []




import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from dpdispatcher.ssh_context import SSHSession, SSHContext
from dpdispatcher.submission import Submission, Task, Resources
from dpdispatcher.lsf import LSF

# edit ssh parameters
ssh_session = SSHSession(
    hostname='127.0.0.1',
    port=22,
    remote_root='/home/dp/dpdispatcher/tests',
    username='debug'
)
# local_root is the relevant path that containing tasks to be submit
ssh_context = SSHContext(local_root='test_lsf_dir', ssh_session=ssh_session)
lsf = LSF(context=ssh_context)

prepend_text = '''
module load cuda/9.2
module load gcc/4.9.4
module load deepmd/1.0
source /home/dp/scripts/avail_gpu.sh
'''

lsf_bsub_dict = {
    'R': "'select[hname != g005]'"
}
resources = Resources(
    number_node=1,
    cpu_per_node=4,
    gpu_per_node=0,
    queue_name="gpu",
    walltime="24:00:00",
    prepend_text=prepend_text,
    append_text="",
    gpu_usage=False,
    gpu_new_syntax=False,
    extra_specification=lsf_bsub_dict,
    group_size=1
)

# task_need_resources has no effect
submission = Submission(
    work_base='0_md',  # the dir containing all of task_work_path
    resources=resources,  # resources above
    forward_common_files=['graph.pb'],  # file to be upload
    backward_common_files=['*.json']  # file to be downloaded
)
task1 = Task(
    command='lmp_mpi_20201029 -i input.lammps',
    task_work_path='bct-1',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps'],
)
task2 = Task(
    command='lmp_mpi_20201029 -i input.lammps',
    task_work_path='bct-2',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps'],
)
task3 = Task(
    command='lmp_mpi_20201029 -i input.lammps',
    task_work_path='bct-3',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps'],
)
task4 = Task(
    command='lmp_mpi_20201029 -i input.lammps',
    task_work_path='bct-4',
    forward_files=['conf.lmp', 'input.lammps'],
    backward_files=['log.lammps'],
)
submission.register_task_list([task1, task2, task3, task4, ])
submission.generate_jobs()
submission.bind_batch(batch=lsf)

submission.run_submission()



import os,getpass,time
from dpdispatcher.Batch import Batch
from dpdispatcher.JobStatus import JobStatus
import datetime

def _default_item(resources, key, value) :
    if key not in resources :
        resources[key] = value


class Shell(Batch) :

    def check_status(self) :
        if self.check_finish_tag():
            return JobStatus.finished
        elif self.check_running():
            return JobStatus.running
        else:
            return JobStatus.terminated
        ## warn: cannont distinguish terminated from unsubmitted.

    def do_submit(self, 
                  job_dirs,
                  cmd,
                  args = None, 
                  res = None,
                  outlog = 'log',
                  errlog = 'err'):
        if res == None:
            res = {}
        script_str = self.sub_script(job_dirs, cmd, args=args, res=res, outlog=outlog, errlog=errlog)
        self.context.write_file(self.sub_script_name, script_str)
        self.proc = self.context.call('cd %s && exec bash %s' % (self.context.remote_root, self.sub_script_name))

    def check_running(self):
        uuid_names = self.context.job_uuid
        ## Check if the uuid.sub is running on remote machine
        cnt = 0
        ret, stdin, stdout, stderr = self.context.block_call("ps aux | grep %s"%uuid_names)
        response_list = stdout.read().decode('utf-8').split("\n")
        for response in response_list:
            if  uuid_names + ".sub" in response:
                return True
        return False

    def default_resources(self, res_) :
        if res_ == None :
            res = {}
        else:
            res = res_
        _default_item(res, 'task_per_node', 1)
        _default_item(res, 'module_list', [])
        _default_item(res, 'module_unload_list', [])
        _default_item(res, 'source_list', [])
        _default_item(res, 'envs', {})
        _default_item(res, 'with_mpi', False)
        _default_item(res, 'cuda_multi_tasks', False)
        _default_item(res, 'allow_failure', False)
        _default_item(res, 'cvasp', False)
        return res

    def sub_script_head(self, resources) :
        envs = resources['envs']
        module_list = resources['module_list']
        module_unload_list = resources['module_unload_list']
        task_per_node = resources['task_per_node']
        source_list = resources['source_list']
        
        ret = ''
        ret += ('#!/bin/bash\n\n')
        # fp.write('set -euo pipefail\n')
        for key in envs.keys() :
            ret += ('export %s=%s\n' % (key, envs[key]))
        ret += ('\n')
        for ii in module_unload_list :
            ret += ('module unload %s\n' % ii)
        ret += ('\n')
        for ii in module_list :
            ret += ('module load %s\n' % ii)
        ret += ('\n')
        for ii in source_list :
            ret += ('source %s\n' % ii)
        ret += ('\n')
        return ret


    def sub_script_cmd(self,
                       cmd,
                       arg,
                       res) :
        try:
            cvasp=res['cvasp']
            fp_max_errors = 3
            try:
                fp_max_errors = res['fp_max_errors']
            except:
                pass
        except:
            cvasp=False

        _cmd = cmd.split('1>')[0].strip()
        if cvasp :
            if res['with_mpi']:
                _cmd = 'python cvasp.py "mpirun -n %d %s %s" %s' % (res['task_per_node'], _cmd, arg, fp_max_errors)
            else :
                _cmd = 'python cvasp.py "%s %s" %s' % (_cmd, arg, fp_max_errors)
        else :
            if res['with_mpi']:
                _cmd = 'mpirun -n %d %s %s' % (res['task_per_node'],  _cmd, arg)
            else :
                _cmd = '%s %s' % (_cmd, arg)
        return _cmd



import os,sys,json,glob,shutil,uuid,time
import unittest
from unittest.mock import MagicMock, patch, PropertyMock

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'tests'
from .context import LocalContext
from .context import PBS
from .context import JobStatus
from .context import LazyLocalContext, LocalContext, SSHContext
from .context import Submission, Job, Task, Resources
from .context import LSF, Slurm, PBS, Shell
from .context import Machine


# print('in', SampleClass.get_sample_empty_submission())

class TestMachineDispatch(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_lazy_local(self):
        machine_dict = {
            'batch_type': 'PBS',
            'context_type': 'LazyLocalContext',
            'local_root':'./'
        }
        machine = Machine.load_from_dict(
            machine_dict=machine_dict
        )
        # pylint: disable=maybe-no-member
        self.assertIsInstance(machine.context, LazyLocalContext)

    def test_local(self):
        machine_dict = {
            'batch_type': 'PBS',
            'context_type': 'LocalContext',
            'local_root': './',
            'remote_root': './'
        }
        machine = Machine.load_from_dict(
            machine_dict=machine_dict
        )
        # pylint: disable=maybe-no-member
        self.assertIsInstance(machine.context, LocalContext)

    def test_ssh(self):
        pass
        # jdata = {
        #     'batch_type': 'pbs',
        #     'context_type': 'ssh',
        #     'hostname': 'localhost',
        #     'local_root': './',
        #     "remote_root" : "/home/fengbo/work_path_dpdispatcher_test/",
        #     "username" : "fengbo"
        # }
        # batch = BatchObject(
        #     jdata=jdata
        # )
        # self.assertIsInstance(batch.context, SSHContext)
    def test_key_err(self):
        # pass
        machine_dict = {}
        with self.assertRaises(KeyError):
            Machine.load_from_dict(machine_dict=machine_dict)

    def test_context_err(self):
        machine_dict = {
            'batch_type': 'PBS',
            'context_type': 'foo'
        }
        with self.assertRaises(KeyError):
            Machine.load_from_dict(machine_dict=machine_dict)

    def test_pbs(self):
        machine_dict = {
            'batch_type': 'PBS',
            'context_type': 'LazyLocalContext',
            'local_root':'./'
        }
        machine = Machine.load_from_dict(
            machine_dict=machine_dict
        )
        self.assertIsInstance(machine, PBS)

    def test_lsf(self):
        machine_dict = {
            'batch_type': 'LSF',
            'context_type': 'LazyLocalContext',
            'local_root':'./'
        }
        machine = Machine.load_from_dict(
            machine_dict=machine_dict
        )
        self.assertIsInstance(machine, LSF)

    def test_slurm(self):
        machine_dict = {
            'batch_type': 'Slurm',
            'context_type': 'LazyLocalContext',
            'local_root':'./'
        }
        machine = Machine.load_from_dict(
            machine_dict=machine_dict
        )
        self.assertIsInstance(machine, Slurm)

    def test_shell(self):
        machine_dict = {
            'batch_type': 'Shell',
            'context_type': 'LazyLocalContext',
            'local_root':'./'
        }
        machine = Machine.load_from_dict(
            machine_dict=machine_dict
        )
        self.assertIsInstance(machine, Shell)


import os,sys,time

from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog


class Batch(object) :
    def __init__ (self,
                  context, 
                  uuid_names = True) :
        self.context = context
        self.uuid_names = uuid_names
        if uuid_names:
            self.upload_tag_name = '%s_tag_upload' % self.context.job_uuid
            self.finish_tag_name = '%s_tag_finished' % self.context.job_uuid
            self.sub_script_name = '%s.sub' % self.context.job_uuid
            self.job_id_name = '%s_job_id' % self.context.job_uuid
        else:
            self.upload_tag_name = 'tag_upload'
            self.finish_tag_name = 'tag_finished'
            self.sub_script_name = 'run.sub'
            self.job_id_name = 'job_id'

    def check_status(self) :
        raise RuntimeError('abstract method check_status should be implemented by derived class')        
        
    def default_resources(self, res) :
        raise RuntimeError('abstract method sub_script_head should be implemented by derived class')        

    def sub_script_head(self, res) :
        raise RuntimeError('abstract method sub_script_head should be implemented by derived class')        

    def sub_script_cmd(self, cmd, res):
        raise RuntimeError('abstract method sub_script_cmd should be implemented by derived class')        

    def do_submit(self,
                  job_dirs,
                  cmd,
                  args = None, 
                  res = None,
                  outlog = 'log',
                  errlog = 'err'):
        '''
        submit a single job, assuming that no job is running there.
        '''
        raise RuntimeError('abstract method check_status should be implemented by derived class')        

    def sub_script(self,
                   job_dirs,
                   cmd,
                   args = None,
                   res  = None,
                   outlog = 'log',
                   errlog = 'err') :
        """
        make submit script

        job_dirs(list):         directories of jobs. size: n_job
        cmd(list):              commands to be executed. size: n_cmd
        args(list of list):     args of commands. size of n_cmd x n_job
                                can be None
        res(dict):              resources available
        outlog(str):            file name for output
        errlog(str):            file name for error
        """
        res = self.default_resources(res)
        ret = self.sub_script_head(res)
        if not isinstance(cmd, list):
            cmd = [cmd]
        if args == None :
            args = []
            for ii in cmd:
                _args = []
                for jj in job_dirs:
                    _args.append('')
                args.append(_args)
        # loop over commands 
        self.cmd_cnt = 0
        try:
            self.manual_cuda_devices = res['manual_cuda_devices']
        except KeyError:
            self.manual_cuda_devices = 0
        try:
            self.manual_cuda_multiplicity = res['manual_cuda_multiplicity']
        except KeyError:
            self.manual_cuda_multiplicity = 1
        for ii in range(len(cmd)):            
            # for one command
            ret += self._sub_script_inner(job_dirs,
                                          cmd[ii],
                                          args[ii],
                                          ii,
                                          res,
                                          outlog=outlog,
                                          errlog=errlog)
        ret += '\ntouch %s\n' % self.finish_tag_name
        return ret

    def submit(self,
               job_dirs,
               cmd,
               args = None,
               res = None,
               restart = False,
               outlog = 'log',
               errlog = 'err'):
        if restart:
            dlog.debug('restart task')
            status = self.check_status()
            if status in [  JobStatus.unsubmitted, JobStatus.unknown, JobStatus.terminated ]:
                dlog.debug('task restart point !!!')
                self.do_submit(job_dirs, cmd, args, res, outlog=outlog, errlog=errlog)
            elif status==JobStatus.waiting:
                dlog.debug('task is waiting')
            elif status==JobStatus.running:
                dlog.debug('task is running')
            elif status==JobStatus.finished:
                dlog.debug('task is finished')
            else:
                raise RuntimeError('unknow job status, must be wrong')
        else:
            dlog.debug('new task')
            self.do_submit(job_dirs, cmd, args, res, outlog=outlog, errlog=errlog)
        if res is None:
            sleep = 0
        else:
            sleep = res.get('submit_wait_time', 0)
        time.sleep(sleep) # For preventing the crash of the tasks while submitting

    def check_finish_tag(self) :
        return self.context.check_file_exists(self.finish_tag_name)

    def _sub_script_inner(self, 
                          job_dirs,
                          cmd,
                          args,
                          idx,
                          res,
                          outlog = 'log',
                          errlog = 'err') :
        ret = ""
        allow_failure = res.get('allow_failure', False)
        for ii,jj in zip(job_dirs, args) :
            ret += 'cd %s\n' % ii
            ret += 'test $? -ne 0 && exit 1\n\n'
            if self.manual_cuda_devices <= 0:
                ret += 'if [ ! -f tag_%d_finished ] ;then\n' % idx
                ret += '  %s 1>> %s 2>> %s \n' % (self.sub_script_cmd(cmd, jj, res), outlog, errlog)
                if res['allow_failure'] is False:
                    ret += '  if test $? -ne 0; then exit 1; else touch tag_%d_finished; fi \n' % idx
                else :
                    ret += '  if test $? -ne 0; then touch tag_failure_%d; fi \n' % idx
                    ret += '  touch tag_%d_finished \n' % idx
                ret += 'fi\n\n'
            else :
                # do not support task-wise restart
                tmp_cmd = ' %s 1>> %s 2>> %s ' % (self.sub_script_cmd(cmd, jj, res), outlog, errlog)
                ret += 'export CUDA_VISIBLE_DEVICES=%d; %s &\n\n' % ((self.cmd_cnt % self.manual_cuda_devices), tmp_cmd)
                self.cmd_cnt += 1
            ret += 'cd %s\n' % self.context.remote_root
            ret += 'test $? -ne 0 && exit 1\n'
            if self.manual_cuda_devices > 0 and self.cmd_cnt % (self.manual_cuda_devices * self.manual_cuda_multiplicity) == 0:
                ret += '\nwait\n\n'
        ret += '\nwait\n\n'
        return ret



import os,getpass,time
from datetime import datetime
from itertools import zip_longest
from dpdispatcher.Batch import Batch 
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog

try:
    import boto3
except ModuleNotFoundError:
    pass
else:
    batch_client = boto3.client('batch')

class AWS(Batch):
    _query_time_interval = 30
    _job_id_map_status = {}
    _jobQueue = ""
    _query_next_allow_time = datetime.now().timestamp()

    @staticmethod
    def map_aws_status_to_dpdisp_status(aws_status):
        map_dict = {'SUBMITTED': JobStatus.waiting,
                'PENDING': JobStatus.waiting,
                'RUNNABLE': JobStatus.waiting,
                'STARTING': JobStatus.waiting,
                'RUNNING': JobStatus.running,
                'SUCCEEDED': JobStatus.finished,
                'FAILED': JobStatus.terminated,
                'UNKNOWN': JobStatus.unknown}
        return map_dict.get(aws_status, JobStatus.unknown)

    @classmethod
    def AWS_check_status(cls, job_id=""):
        """
        to aviod query jobStatus too often, set a time interval
        query_dict example:
        {job_id: JobStatus}

        {'40fb24b2-d0ca-4443-8e3a-c0906ea03622': <JobStatus.running: 3>,
         '41bda50c-0a23-4372-806c-87d16a680d85': <JobStatus.waiting: 2>}
           
        """
        query_dict ={}
        if datetime.now().timestamp() > cls._query_next_allow_time:
            cls._query_next_allow_time=datetime.now().timestamp()+cls._query_time_interval
            for status in ['SUBMITTED', 'PENDING', 'RUNNABLE', 'STARTING', 'RUNNING','SUCCEEDED', 'FAILED']:
                nextToken = ''
                while nextToken is not None:
                    status_response = batch_client.list_jobs(jobQueue=cls._jobQueue, jobStatus=status, maxResults=100, nextToken=nextToken)
                    status_list=status_response.get('jobSummaryList')
                    nextToken = status_response.get('nextToken', None)
                    for job_dict in status_list:
                        cls._job_id_map_status.update({job_dict['jobId']: cls.map_aws_status_to_dpdisp_status(job_dict['status'])})
            dlog.debug('20000:_map: %s' %(cls._job_id_map_status))
        dlog.debug('62000:job_id:%s, _query: %s, _map: %s' %(job_id, query_dict, cls._job_id_map_status))
        if job_id:
            return cls._job_id_map_status.get(job_id)
                    
        return cls._job_id_map_status

    @property
    def job_id(self):
        try:
            self._job_id
        except AttributeError:
            if self.context.check_file_exists(self.job_id_name):
                self._job_id = self.context.read_file(self.job_id_name)
                response_list = batch_client.describe_jobs(jobs=[self._job_id]).get('jobs')
                try:
                    response = response_list[0]
                    jobQueue = response['jobQueue']
                except IndexError:
                    pass
                else:
                    self.job_id = (response, jobQueue)
                    return self._job_id
            dlog.debug("50000, self._job_id:%s,_Queue:%s,_map:%s,"%(self._job_id, self.__class__._jobQueue, self.__class__._job_id_map_status  ))       
            return ""
        return self._job_id

    @job_id.setter
    def job_id(self, values):
        response, jobQueue = values
        self._job_id = response['jobId']
        self._job_name = response['jobName']
        self.__class__._jobQueue = jobQueue
        self.__class__._job_id_map_status[self._job_id] = self.map_aws_status_to_dpdisp_status(response.get('status', 'SUBMITTED'))
        self.context.write_file(self.job_id_name, self._job_id)
        dlog.debug("15000, _job_id:%s, _job_name:%s, _map:%s, _Queue:%s" % (self._job_id, self._job_name, self.__class__._job_id_map_status, self.__class__._jobQueue))

    def check_status(self):
        return self.__class__.AWS_check_status(job_id=self.job_id)
    
    def sub_script(self, job_dirs, cmd, args, res, outlog, errlog):
        if args is None:
            args=[]
        multi_command = ""
        for job_dir in job_dirs:
            for idx,t in enumerate(zip_longest(cmd, args, fillvalue='')):
                c_str =  f"cd {self.context.remote_root}/{job_dir} && ( test -f tag_{idx}_finished || ( ({t[0]} {t[1]} && touch tag_{idx}_finished  2>>{errlog} || exit 52 ) | tee -a {outlog}) ) || exit 51;"
                multi_command += c_str
        multi_command +="exit 0;"
        dlog.debug("10000, %s" % multi_command)
        return multi_command
        
    def default_resources(self, res):
        if res == None:
            res = {}
        else:
            # res.setdefault(jobDefinition)
            res.setdefault('cpu_num', 32)
            res.setdefault('memory_size', 120000)
            res.setdefault('jobQueue', 'deepmd_m5_v1_7')
        return res
    
    def do_submit(self,
            job_dirs,
            cmd,
            args = None,
            res = None,
            outlog = 'log',
            errlog = 'err'):

        res = self.default_resources(res)
        dlog.debug("2000, params=(%s, %s, %s, %s, %s, %s, )" % (job_dirs, cmd, args, res,  outlog, errlog ))
        dlog.debug('2200, self.context.remote_root: %s , self.context.local_root: %s' % (self.context.remote_root, self.context.local_root))
        # concreate_command = 
        script_str = self.sub_script(job_dirs, cmd, args=args, res=res, outlog=outlog, errlog=errlog)
        dlog.debug('2300, script_str: %s, self.sub_script_name: %s' % (script_str, self.sub_script_name))
        """
        jobName example:
        home-ec2-user-Ag_init-run_gen-iter_000000-01_model_devi-task_000_000048
        """
        jobName = os.path.join(self.context.remote_root,job_dirs.pop())[1:].replace('/','-').replace('.','_')
        jobName += ("_" + str(self.context.job_uuid))
        response = batch_client.submit_job(jobName=jobName, 
                jobQueue=res['jobQueue'], 
                jobDefinition=res['jobDefinition'],
                parameters={'task_command':script_str},
                containerOverrides={'vcpus':res['cpu_num'], 'memory':res['memory_size']})
        dlog.debug('4000, response:%s' % response)
        self.job_id = (response, res['jobQueue'])




import os,sys,time,random,uuid,json

from dpdispatcher.JobStatus import JobStatus
from dpdispatcher import dlog
from dpdispatcher.base_context import BaseContext

script_template="""\
{script_header}
{script_custom_flags}
{script_env}
{script_command}
{script_end}
"""

script_env_template="""
export REMOTE_ROOT={remote_root}
test $? -ne 0 && exit 1
{source_files_part}
"""

script_command_template="""
cd $REMOTE_ROOT
cd {task_work_path}
test $? -ne 0 && exit 1
if [ ! -f {task_tag_finished} ] ;then
  {command_env} {command} {log_err_part}
  if test $? -eq 0; then touch {task_tag_finished}; fi
fi &
"""

script_end_template="""
cd $REMOTE_ROOT
test $? -ne 0 && exit 1

wait

touch {job_tag_finished}
"""


class Machine(object):
    subclasses_dict = {}
    def __init__ (self,
                context):
        self.context = context
        # self.uuid_names = uuid_names
        # self.upload_tag_name = '%s_job_tag_upload' % self.context.job_uuid
        # self.finish_tag_name = '%s_job_tag_finished' % self.context.job_uuid
        # self.sub_script_name = '%s.sub' % self.context.job_uuid
        # self.job_id_name = '%s_job_id' % self.context.job_uuid

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.subclasses_dict[cls.__name__]=cls
        # cls.subclasses.append(cls)

    @classmethod
    def load_from_json(cls, json_path):
        with open(json_path, 'r') as f:
            machine_dict = json.load(f)
        machine = cls.load_from_dict(machine_dict=machine_dict)
        return machine

    @classmethod
    def load_from_dict(cls, machine_dict):
        batch_type = machine_dict['batch_type']
        # print("debug777:batch_class", cls.subclasses_dict, batch_type)
        machine_class = cls.subclasses_dict[batch_type]
        context = BaseContext.load_from_dict(machine_dict)
        machine = machine_class(context=context)
        return machine

    def check_status(self, job) :
        raise NotImplementedError('abstract method check_status should be implemented by derived class')        
        
    def default_resources(self, res) :
        raise NotImplementedError('abstract method sub_script_head should be implemented by derived class')        

    def sub_script_head(self, res) :
        raise NotImplementedError('abstract method sub_script_head should be implemented by derived class')        

    def sub_script_cmd(self, res):
        raise NotImplementedError('abstract method sub_script_cmd should be implemented by derived class')        

    def do_submit(self, job):
        '''
        submit a single job, assuming that no job is running there.
        '''
        raise NotImplementedError('abstract method do_submit should be implemented by derived class')        

    def gen_script(self, job):
        script_header = self.gen_script_header(job)
        script_custom_flags = self.gen_script_custom_flags_lines(job)
        script_env = self.gen_script_env(job)
        script_command = self.gen_script_command(job)
        script_end = self.gen_script_end(job)
        script = script_template.format(
            script_header=script_header,
            script_custom_flags=script_custom_flags,
            script_env=script_env,
            script_command=script_command,
            script_end=script_end
        )
        return script

    def check_if_recover(self, submission):
        submission_hash = submission.submission_hash
        submission_file_name = "{submission_hash}.json".format(submission_hash=submission_hash)
        if_recover = self.context.check_file_exists(submission_file_name)
        return if_recover

    def check_finish_tag(self, **kwargs):
        raise NotImplementedError('abstract method check_finish_tag should be implemented by derived class')        

    def gen_script_header(self, job):
        raise NotImplementedError('abstract method gen_script_header should be implemented by derived class')

    def gen_script_custom_flags_lines(self, job):
        custom_flags_lines = ""
        custom_flags = job.resources.custom_flags
        for ii in custom_flags:
            line = ii + '\n'
            custom_flags_lines += line 
        return custom_flags_lines

    def gen_script_env(self, job):
        source_files_part = ""
        source_list = job.resources.source_list
        for ii in source_list:
            line = f"source {ii}\n"
            source_files_part += line

        script_env = script_env_template.format(
            remote_root=self.context.remote_root,
            source_files_part=source_files_part)
        return script_env

    def gen_script_command(self, job):
        script_command = ""
        resources = job.resources
        # in_para_task_num = 0
        for task in job.job_task_list:
            command_env = ""
            command_env += self.gen_command_env_cuda_devices(resources=resources)

            task_tag_finished = task.task_hash + '_task_tag_finished'

            log_err_part = ""
            if task.outlog is not None:
                log_err_part += f"1>>{task.outlog} "
            if task.errlog is not None:
                log_err_part += f"2>>{task.errlog} "

            single_script_command = script_command_template.format(command_env=command_env, 
                task_work_path=task.task_work_path, command=task.command, task_tag_finished=task_tag_finished,
                log_err_part=log_err_part)
            script_command += single_script_command

            script_command += self.gen_script_wait(resources=resources)
        return script_command

    def gen_script_end(self, job):
        job_tag_finished = job.job_hash + '_job_tag_finished'
        script_end = script_end_template.format(job_tag_finished=job_tag_finished)
        return script_end

    def gen_script_wait(self, resources):
        # if not resources.strategy.get('if_cuda_multi_devices', None):
        #     return "wait \n"
        para_deg = resources.para_deg
        resources.task_in_para += 1
        # task_need_gpus = task.task_need_gpus
        if resources.task_in_para >= para_deg:
            # pbs_script_command += pbs_script_wait
            resources.task_in_para = 0
            if resources.strategy['if_cuda_multi_devices'] is True:
                resources.gpu_in_use += 1
                if resources.gpu_in_use % resources.gpu_per_node == 0:
                    return "wait \n"
                else:
                    return ""
            return "wait \n"
        return ""

    def gen_command_env_cuda_devices(self, resources):
        # task_need_resources = task.task_need_resources
        # task_need_gpus = task_need_resources.get('task_need_gpus', 1)
        command_env = ""
        # gpu_number = resources.gpu_per_node
        # resources.gpu_in_use = 0

        if resources.strategy['if_cuda_multi_devices'] is True:
            if resources.gpu_per_node == 0:
                raise RuntimeError("resources.gpu_per_node can not be 0")
            gpu_index = resources.gpu_in_use % resources.gpu_per_node
            command_env+=f"export CUDA_VISIBLE_DEVICES={gpu_index};"
            # for ii in list_CUDA_VISIBLE_DEVICES:
            #     command_env+="{ii},".format(ii=ii) 
        return command_env


